@@ -1,7 +1,7 @@
 /* dld -- dynamic link/unlink editor for C
    Copyright (C) 1990 by W. Wilson Ho.
 
-   Version 3.2.3
+   Version 3.2.6
    The author can be reached electronically by how@cs.ucdavis.edu or
    through physical mail at:
 
@@ -43,7 +43,7 @@
 /*
  * This part is taken out after the elimination of use of alloca().
  */
-/* 
+/*
  * Alloca include.
  */
 /* If compiled with GNU C, use the built-in alloca */
@@ -68,7 +68,7 @@ extern long _initial_stack = - 100 * 1024L;
 #endif /* atarist */
 
 /* Ordinary 4.3 bsd lacks these macros in a.out.h */
-    
+
 #ifndef N_TXTADDR
 #ifdef vax
 #define N_TXTADDR(x) 0
@@ -86,8 +86,8 @@ extern long _initial_stack = - 100 * 1024L;
 #endif
 #ifdef sequent
 #define N_DATADDR(x) \
-        (((x).a_magic==OMAGIC)? (N_TXTADDR(x)+(x).a_text) \
-        : (page_size+(((x).a_text-1) & ~(page_size-1))))
+	(((x).a_magic==OMAGIC)? (N_TXTADDR(x)+(x).a_text) \
+	: (page_size+(((x).a_text-1) & ~(page_size-1))))
 #endif
 #endif
 
@@ -159,17 +159,17 @@ extern long _initial_stack = - 100 * 1024L;
 #if defined(sun) && defined(sparc)
 /* Sparc (Sun 4) macros */
 #undef relocation_info
-#define relocation_info	                reloc_info_sparc
-#define RELOC_ADDRESS(r)		((r)->r_address)                 
-#define RELOC_EXTERN_P(r)               ((r)->r_extern)      
-#define RELOC_SYMBOL(r)                 ((r)->r_index)   
+#define relocation_info			reloc_info_sparc
+#define RELOC_ADDRESS(r)		((r)->r_address)
+#define RELOC_EXTERN_P(r)		((r)->r_extern)
+#define RELOC_SYMBOL(r)			((r)->r_index)
 #define RELOC_MEMORY_SUB_P(r)		0
-#define RELOC_ADD_EXTRA(r)              ((r)->r_addend)       
-#define RELOC_PCREL_P(r)             \
-        ((r)->r_type >= RELOC_DISP8 && (r)->r_type <= RELOC_WDISP22)
-#define RELOC_VALUE_RIGHTSHIFT(r)       (reloc_target_rightshift[(r)->r_type])
-#define RELOC_TARGET_SIZE(r)            (reloc_target_size[(r)->r_type])
-#define RELOC_TARGET_BITSIZE(r)         (reloc_target_bitsize[(r)->r_type])
+#define RELOC_ADD_EXTRA(r)		((r)->r_addend)
+#define RELOC_PCREL_P(r)		\
+	((r)->r_type >= RELOC_DISP8 && (r)->r_type <= RELOC_WDISP22)
+#define RELOC_VALUE_RIGHTSHIFT(r)	(reloc_target_rightshift[(r)->r_type])
+#define RELOC_TARGET_SIZE(r)		(reloc_target_size[(r)->r_type])
+#define RELOC_TARGET_BITSIZE(r)		(reloc_target_bitsize[(r)->r_type])
 
 /* Note that these are very dependent on the order of the enums in
    enum reloc_type (in a.out.h); if they change the following must be
@@ -286,10 +286,11 @@ register int errno;
 
 /* Like malloc but get fatal error if memory is exhausted.  */
 
+int
 _dld_malloc (size)
 int size;
 {
-    register int result = malloc (size);
+    register int result = (int) malloc (size);
     if (!result)
 	fatal (DLD_ENOMEMORY);
     return result;
@@ -305,31 +306,31 @@ char *s1, *s2, *s3;
 {
     register int len1 = strlen (s1), len2 = strlen (s2), len3 = strlen (s3);
     register char *result = (char *) _dld_malloc (len1 + len2 + len3 + 1);
-    
+
     strcpy (result, s1);
     strcpy (result + len1, s2);
     strcpy (result + len1 + len2, s3);
     result[len1 + len2 + len3] = 0;
-    
+
     return result;
 } /* concat */
 
 
 /* Add a new entry to the file chain if it is not already there.
    Return 0 if no actual insertion is needed, otherwise, return 1. */
-static
+static int
 insert_entry (head, entry)
 struct file_chain **head;
 register struct file_entry *entry;
 {
     register struct file_chain *fc = *head;
-    
+
     while (fc)  {
-        if (fc->entry == entry)
-            return 0;
-        else fc = fc->next;
+	if (fc->entry == entry)
+	    return 0;
+	else fc = fc->next;
     }
-    
+
     fc = (struct file_chain *) _dld_malloc (sizeof (struct file_chain));
 
     fc->next = *head;
@@ -358,26 +359,28 @@ file_open (entry)
 register struct file_entry *entry;
 {
     register int desc;
-    
+
     if (entry == 0) fatal (DLD_ENOFILE);
-    
+
     if (entry->superfile)
 	return file_open (entry->superfile);
-    
+
     if (entry == input_file)
 	return input_desc;
-    
+
     if (input_file) file_close ();
-    
+
     desc = open (entry->filename, O_RDONLY, 0);
-    
+
     if (desc > 0) {
 	input_file = entry;
 	input_desc = desc;
 	return desc;
     }
-    
+
     fatal (DLD_ENOFILE);
+
+    return(1);  /* To supress warning message */
 } /* file_open */
 
 /* Medium-level input routines for rel files.  */
@@ -393,14 +396,14 @@ register struct file_entry *entry;
 {
     register int len;
     struct exec *loc = &entry->header;
-    
+
     lseek (desc, entry->starting_offset, 0);
     len = read (desc, loc, sizeof (struct exec));
     if (len != sizeof (struct exec))
 	fatal (DLD_EBADHEADER);
     if (N_BADMAG (*loc))
 	fatal (DLD_EBADMAGIC);
-    
+
     entry->header_read_flag = 1;
 } /* read_header */
 
@@ -415,7 +418,7 @@ register struct file_entry *entry;
     register int len;
     struct exec *loc = &entry->header;
     struct aexec file_header;
-    
+
     lseek (desc, 0L, 0);
     len = read (desc, &file_header, sizeof (struct aexec));
     if (len != sizeof (struct aexec))
@@ -446,15 +449,15 @@ struct file_entry *entry;
 int desc;
 {
     int str_size;
-    
+
     if (!entry->header_read_flag)
 	read_header (desc, entry);
-    
+
     if (entry->header.a_syms <= 0)
 	fatal (DLD_ENOSYMBOLS);
-    
+
     entry->symbols = (struct nlist *) _dld_malloc (entry->header.a_syms);
-    
+
     lseek (desc, N_SYMOFF (entry->header) + entry->starting_offset, 0);
     if (entry->header.a_syms !=
 	read (desc, entry->symbols, entry->header.a_syms)) {
@@ -462,14 +465,14 @@ int desc;
 	entry->symbols = 0;
 	fatal (DLD_ENOSYMBOLS);
     }
-    
+
     lseek (desc, N_STROFF (entry->header) + entry->starting_offset, 0);
     if (sizeof str_size != read (desc, &str_size, sizeof str_size)) {
 	free (entry->symbols);
 	entry->symbols = 0;
 	fatal (DLD_ENOSTRINGS);
     }
-    
+
     entry->string_size = str_size;
 } /* read_entry_symboles */
 
@@ -484,7 +487,7 @@ int desc;
 {
     if (!entry->header_read_flag)
 	read_header (desc, entry);
-    
+
     entry->strings = (char *) _dld_malloc (entry->string_size);
 
     lseek (desc, N_STROFF (entry->header) + entry->starting_offset, 0);
@@ -494,7 +497,7 @@ int desc;
 	entry->strings = 0;
 	fatal (DLD_ENOSTRINGS);
     }
-    
+
 } /* read_entry_strings */
 
 
@@ -504,6 +507,7 @@ int desc;
    RELOC_SIZE is its length in bytes.
    If everything is ok, return the number of external relocation entry.
    Otherwise, return -1. */
+int
 reloc_info_ok (data_size, reloc_info, reloc_size, sym_size)
 register int data_size;
 register struct relocation_info *reloc_info;
@@ -514,10 +518,10 @@ register int sym_size;
     register struct relocation_info *end = p +
 	reloc_size / sizeof (struct relocation_info);
     register int extern_count = 0;  /* number of external relocation */
-    
+
     for (; p < end; p++) {
 	register int symbolnum = RELOC_SYMBOL(p);
-	
+
 	if (RELOC_ADDRESS(p) >= data_size)
 	    return -1;
 	if (RELOC_EXTERN_P(p)) {
@@ -531,8 +535,8 @@ register int sym_size;
 
 	if (RELOC_TARGET_SIZE(p) > 2) return -1;
     }
-    return extern_count++;
-    
+    return extern_count++;     /*  this postincrement does nothing */
+			       /*  but preincrement causes bug     */
 } /* reloc_info_ok */
 
 /* symbol table management */
@@ -545,12 +549,12 @@ char *key;
 {
     register char *cp;
     register int k;
-    
+
     cp = key;
     k = 0;
     while (*cp)
 	k = (((k << 1) + (k >> 14)) ^ (*cp++)) & 0x3fff;
-    
+
     return k;
 } /* hash_string */
 
@@ -564,29 +568,29 @@ char *key;
 {
     register int hashval;
     register symbol *bp;
-    
+
     /* Determine the proper bucket.  */
-    
+
     hashval = hash_string (key) % TABSIZE;
-    
+
     /* Search the bucket.  */
-    
+
     for (bp = _dld_symtab[hashval]; bp; bp = bp->link)
 	if (! strcmp (key, bp->name))
 	    return bp;
-    
+
     /* Nothing was found; create a new symbol table entry.  */
-    
+
     bp = (symbol *) _dld_malloc (sizeof (symbol));
     bzero (bp, sizeof (symbol));
     bp->name = (char *) _dld_malloc (strlen (key) + 1);
     strcpy (bp->name, key);
-    
+
     /* Add the entry to the bucket.  */
-    
+
     bp->link = _dld_symtab[hashval];
     _dld_symtab[hashval] = bp;
-    
+
     return bp;
 } /* _dld_getsym */
 
@@ -599,17 +603,17 @@ char *key;
 {
     register int hashval;
     register symbol *bp;
-    
+
     /* Determine which bucket.  */
-    
+
     hashval = hash_string (key) % TABSIZE;
-    
+
     /* Search the bucket.  */
-    
+
     for (bp = _dld_symtab[hashval]; bp; bp = bp->link)
 	if (! strcmp (key, bp->name))
 	    return bp;
-    
+
     return 0;
 } /* _dld_getsym_soft */
 
@@ -652,32 +656,36 @@ char *name;
     int olddef = sp->defined;
 
     nlist_p->n_un.n_name = (char *) sp;
-    
+
     sp->referenced = 1;
 
     /* common definition */
     if (common && !olddef)
+#ifndef linux
 	type = N_COMM | N_EXT;
+#else
+	type = N_TYPE | N_EXT;
+#endif
 
     if (type != (N_UNDF | N_EXT)) {
 	/* definition of a symbol */
-	
+
 	if (olddef)
 	    fatal (DLD_EMULTDEFS);
 	else {
 	    sp->defined = type;
 	    sp->value = common ? (long) _dld_malloc (nlist_p->n_value) :
-		nlist_p->n_value; 
-	    if (common) bzero (sp->value, nlist_p->n_value);
+		nlist_p->n_value;
+	    if (common) bzero ((void *)sp->value, nlist_p->n_value);
 	    sp->defined_by = entry;
 	}
 
 	if (oldref && !olddef) {
 	    register struct file_chain *p = sp->referenced_by;
 	    register struct file_chain *prev = 0;
-	    
+
 	    dld_undefined_sym_count--;
-	    
+
 	    /* We do not set up references link to and from the dummy
 	       entry, nor do we touch the ref_count of it.
 	       Furthermore, if a symbol referenced by the dummy entry is
@@ -703,7 +711,7 @@ char *name;
 		    del_link_list_elt (sp->referenced_by, prev, p, next);
 		    continue;
 		} else p->entry->undefined_symbol_count--;
-		
+
 		prev = p;
 		p = p->next;
 	    }
@@ -719,7 +727,7 @@ char *name;
 		sp->defined_by->ref_count++;
 		sp->defined_by->already_unlink = 0;
 	    }
-	
+
 	if (entry != _dld_dummy_entry) {
 	    if (insert_entry (&(sp->referenced_by), entry) && !sp->defined)
 		entry->undefined_symbol_count++;
@@ -727,7 +735,7 @@ char *name;
 	    if (!sp->defined)
 		insert_entry (&(sp->referenced_by), entry);
 	}
-	
+
 	if (!oldref) {
 	    dld_undefined_sym_count++;
 	    sp->defined = 0;
@@ -745,10 +753,33 @@ struct file_entry *entry;
 {
     register struct nlist *p,
 	*end = entry->symbols + entry->header.a_syms / sizeof (struct nlist);
-    
+
     for (p = entry->symbols; p < end; p++)
+#ifdef linux
+	if (p->n_type & N_EXT) {
+	    if (p->n_type == (N_INDR | N_EXT)) {
+	      symbol *sp;
+
+	      /* Enter an indirect symbol and the following undefined symbol
+		 as it's reference (see ".../linux/a.out.h") */
+
+	      /* First enter the referenced symbol */
+	      _dld_enter_global_ref (entry, p+1, (p+1)->n_un.n_strx + entry->strings);
+	      /* Get the address of the corresponding 'struct glosym' */
+	      sp = (symbol *)((p+1)->n_un.n_name);
+	      /* Now enter the indirect symbol (alias) */
+	      _dld_enter_global_ref (entry, p, p->n_un.n_strx + entry->strings);
+	      /* Make 'indirect' point to the referenced symbol */
+	      ((symbol *)(p->n_un.n_name))->indirect = sp;
+	      p++;
+	    }
+	    else
+	      _dld_enter_global_ref (entry, p, p->n_un.n_strx + entry->strings);
+	}
+#else
 	if (p->n_type & N_EXT)
 	    _dld_enter_global_ref (entry, p, p->n_un.n_strx + entry->strings);
+#endif
 }
 
 /* remove all symbols that are no longer needed */
@@ -766,7 +797,7 @@ cleanup_symtab ()
 	    register struct file_chain *prev = 0;
 	    register int obsolete = (sp->defined_by &&
 				     sp->defined_by->ref_count == 0);
-	    
+
 	    while (p)
 		if (p->entry->ref_count == 0) {
 		    del_link_list_elt (sp->referenced_by, prev, p, next);
@@ -779,8 +810,12 @@ cleanup_symtab ()
 		}
 
 	    if (obsolete) {
+#ifndef linux
 		if (sp->defined == (N_COMM | N_EXT) && sp->value)
-		    free (sp->value);
+#else
+		if (sp->defined == (N_TYPE | N_EXT) && sp->value)
+#endif
+		    free ((void *)sp->value);
 		if (sp->referenced_by == 0) {
 		    free (sp->name);
 		    del_link_list_elt (_dld_symtab[i], prev_sp, sp, link);
@@ -796,8 +831,12 @@ cleanup_symtab ()
 	    if (sp->defined == 0 && sp->referenced_by == 0) {
 		dld_undefined_sym_count--;
 		free (sp->name);
+#ifndef linux
 		if (sp->defined == (N_COMM | N_EXT) && sp->value)
-		    free (sp->value);
+#else
+		if (sp->defined == (N_TYPE | N_EXT) && sp->value)
+#endif
+		    free ((void *)sp->value);
 		del_link_list_elt (_dld_symtab[i], prev_sp, sp, link);
 		continue;
 	    }
@@ -828,23 +867,27 @@ struct file_entry *entry;
     int member_length;
     register char *name;
     register struct file_entry *subentry;
-    
+
     if (!dld_undefined_sym_count) return;
-    
+
     /* Examine its first member, which starts SARMAG bytes in.  */
     subentry = decode_library_subfile (desc, entry, SARMAG, &member_length);
     if (!subentry) return;
-    
+
     name = subentry->filename;
     free (subentry);
-    
+
     /* Search via __.SYMDEF if that exists, else linearly.  */
-    
+
+#if defined(linux) && ((__GNUC__ < 2) || ((__GNUC__ == 2) && (__GNUC_MINOR__ <= 5)))
+    if (!strcmp (name, "__.SYMDEF/"))
+#else
     if (!strcmp (name, "__.SYMDEF"))
+#endif
 	symdef_library (desc, entry, member_length);
     else
 	linear_library (desc, entry);
-    
+
     free (name);
 } /* search_library */
 
@@ -866,39 +909,39 @@ int *length_loc;
     register char *name;
     struct ar_hdr hdr1;
     register struct file_entry *subentry;
-    
+
     lseek (desc, subfile_offset, 0);
-    
+
     bytes_read = read (desc, &hdr1, sizeof hdr1);
     if (!bytes_read)
 	return 0;		/* end of archive */
-    
+
     if (sizeof hdr1 != bytes_read)
 	fatal (DLD_EBADLIBRARY);
-    
+
     if (sscanf (hdr1.ar_size, "%d", &member_length) != 1)
 	fatal (DLD_EBADLIBRARY);
-    
+
     subentry = (struct file_entry *) _dld_malloc (sizeof (struct file_entry));
     bzero (subentry, sizeof (struct file_entry));
-    
+
     for (namelen = 0;
 	 namelen < sizeof hdr1.ar_name
 	 && hdr1.ar_name[namelen] != 0 && hdr1.ar_name[namelen] != ' ';
 	 namelen++);
-    
+
     name = (char *) _dld_malloc (namelen+1);
     strncpy (name, hdr1.ar_name, namelen);
     name[namelen] = 0;
-    
+
     subentry->filename = name;
     subentry->local_sym_name = name;
     subentry->starting_offset = subfile_offset + sizeof hdr1;
     subentry->superfile = library_entry;
     subentry->total_size = member_length;
-    
+
     (*length_loc) = member_length;
-    
+
     return subentry;
 } /* decode_library_subfile */
 
@@ -925,7 +968,7 @@ int member_length;
     register int i;
     struct file_entry *prev = 0;
     int prev_offset = 0;
-    
+
     bytes_read = read (desc, symdef_data, member_length);
     if (bytes_read != member_length) {
 	free (symdef_data);
@@ -936,81 +979,81 @@ int member_length;
        entry->strings so that if fatal() is called, this memory block will
        be freed by clean_up().  If no error occurs, it is freed before this
        function returns.  I use entry->strings just for convenience and
-       don't want to add an extra entry in struct file_entry.  
+       don't want to add an extra entry in struct file_entry.
        Originally storage for symdef_data was
        allocated by alloca(), but I want to avoid the use of that. */
 
     entry->strings = (char *) symdef_data;
-    
+
     number_of_symdefs = *symdef_data / sizeof (struct symdef);
     if (number_of_symdefs < 0 ||
 	number_of_symdefs * sizeof (struct symdef) + 2 * sizeof (int) >=
 	member_length)
 	fatal (DLD_EBADLIBRARY);
-    
+
     symdef_base = (struct symdef *) (symdef_data + 1);
     length_of_strings = *(int *) (symdef_base + number_of_symdefs);
-    
+
     if (length_of_strings < 0
 	|| number_of_symdefs * sizeof (struct symdef) + length_of_strings
 	+ 2 * sizeof (int) != member_length)
 	fatal (DLD_EBADLIBRARY);
-    
+
     sym_name_base = sizeof (int) + (char *) (symdef_base + number_of_symdefs);
-    
+
     /* Check all the string indexes for validity.  */
-    
+
     for (i = 0; i < number_of_symdefs; i++) {
 	register int index = symdef_base[i].symbol_name_string_index;
 	if (index < 0 || index >= length_of_strings
 	    || (index && *(sym_name_base + index - 1)))
 	    fatal (DLD_EBADLIBRARY);
     }
-    
+
     /* Search the symdef data for members to load.
        Do this until one whole pass finds nothing to load.  */
-    
+
     not_finished = 1;
     while (not_finished) {
 	not_finished = 0;
-	    
+
 	/* Scan all the symbols mentioned in the symdef for ones that we need.
 	   Load the library members that contain such symbols.  */
-	    
+
 	for (i = 0; i < number_of_symdefs && dld_undefined_sym_count; i++)
 	    if (symdef_base[i].symbol_name_string_index >= 0) {
 		register symbol *sp;
-			
+
 		sp = _dld_getsym_soft (sym_name_base +
 				       symdef_base[i].symbol_name_string_index);
-			
+
 		/* If we find a symbol that appears to be needed,
 		   think carefully about the archive member that
 		   the symbol is in.  */
-			
+
 		if (sp && sp->referenced && !sp->defined) {
 		    int junk;
 		    register int j;
 		    register int offset = symdef_base[i].library_member_offset;
 		    struct file_entry *subentry;
-				
+
 		    /* Don't think carefully about any archive member
 		       more than once in a given pass.  */
-		    
+
 		    if (prev_offset == offset)
 			continue;
 		    prev_offset = offset;
-		    
+
 		    /* Read the symbol table of the archive member.  */
-		    
+
 		    subentry = decode_library_subfile (desc, entry, offset, &junk);
 		    if (prev)
 			prev->chain = subentry;
 		    else entry->subfiles = subentry;
-				
+
 		    read_entry_symbols (desc, subentry);
 		    read_entry_strings (desc, subentry);
-				
+
 		    /* Now scan the symbol table and decide whether to load.  */
 		    if (!subfile_wanted_p (subentry)) {
 			if (prev)
@@ -1024,9 +1067,9 @@ int member_length;
 			/* This member is needed; load it.
 			   Since we are loading something on this pass,
 			   we must make another pass through the symdef data.  */
-			
+
 			not_finished = 1;
-					
+
 			/* all library members are considered unlinked
 			   already so that they will be garbage collected
 			   whenever no other modules reference them. */
@@ -1035,14 +1078,14 @@ int member_length;
 			enter_file_symbols (subentry);
 			free (subentry->strings);
 			subentry->strings = 0;
-					
+
 			read_text_and_data (desc, subentry);
-			
+
 			prev = subentry;
-					
+
 			/* Clear out this member's symbols from the symdef data
 			   so that following passes won't waste time on them.  */
-					
+
 			for (j = 0; j < number_of_symdefs; j++) {
 			    if (symdef_base[j].library_member_offset == offset)
 				symdef_base[j].symbol_name_string_index = -1;
@@ -1055,7 +1098,7 @@ int member_length;
     /* refer to "Ugly kludge" above. */
     free (entry->strings);
     entry->strings = 0;
-} /* symdef_library */ 
+} /* symdef_library */
 
 
 /* Search a library that has no __.SYMDEF.
@@ -1069,18 +1112,18 @@ struct file_entry *entry;
 {
     register struct file_entry *prev = 0;
     register int this_subfile_offset = SARMAG;
-    
+
     while (dld_undefined_sym_count) {
 	int member_length;
 	register struct file_entry *subentry;
-	    
+
 	subentry = decode_library_subfile (desc, entry, this_subfile_offset, &member_length);
-	    
+
 	if (!subentry) return;
-	    
+
 	read_entry_symbols (desc, subentry);
 	read_entry_strings (desc, subentry);
-	    
+
 	if (!subfile_wanted_p (subentry)) {
 	    free (subentry->filename);
 	    free (subentry->symbols);
@@ -1091,15 +1134,15 @@ struct file_entry *entry;
 	    enter_file_symbols (subentry);
 	    free (subentry->strings);
 	    subentry->strings = 0;
-		    
+
 	    read_text_and_data (desc, subentry);
-	    
+
 	    if (prev)
 		prev->chain = subentry;
 	    else entry->subfiles = subentry;
 	    prev = subentry;
 	}
-	    
+
 	this_subfile_offset += member_length + sizeof (struct ar_hdr);
 	if (this_subfile_offset & 1) this_subfile_offset++;
     }
@@ -1117,28 +1160,28 @@ struct file_entry *entry;
     register struct nlist *p;
     register struct nlist *end
 	= entry->symbols + entry->header.a_syms / sizeof (struct nlist);
-    
+
     for (p = entry->symbols; p < end; p++) {
 	register int type = p->n_type;
-	    
+
 	if (type & N_EXT && (type != (N_UNDF | N_EXT) || p->n_value)) {
 	    register char *name = p->n_un.n_strx + entry->strings;
 	    register symbol *sp = _dld_getsym_soft (name);
-		    
+
 	    /* If this symbol has not been hashed, we can't be looking
 	       for it. */
-		    
+
 	    if (!sp) continue;
-		    
+
 	    /* We are looking for definition of external symbols or common
 	       blocks that have not been defined but have been referenced */
-	    
+
 	    if (sp->referenced && !sp->defined &&
 		(type != (N_UNDF | N_EXT) || p->n_value))
 		return 1;
 	}
     }
-    
+
     return 0;
 } /* subfile_wanted_p */
 
@@ -1156,16 +1199,16 @@ register struct file_entry *entry;
     data_relocation = entry->data_start_address - entry->header.a_text;
     bss_relocation = entry->bss_start_address - entry->header.a_text -
 	entry->header.a_data;
-	
+
     for (p = entry->symbols; p < end; p++) {
 	/* If this belongs to a section,
 	   update it by the section's start address */
-		
+
 	register int type = p->n_type;
 	register symbol *sp;
-	
+
 	if ((type & N_EXT) == 0) continue;
-	
+
 	sp = (symbol *) p->n_un.n_name;
 	if (type == (N_TEXT | N_EXT))
 	    sp->value += text_relocation;
@@ -1205,7 +1248,7 @@ int reloc_size;
     int data_relocation = entry->data_start_address - entry->header.a_text;
     int bss_relocation
 	= entry->bss_start_address - entry->header.a_text - entry->header.a_data;
-    
+
     for (; p < end; p++) {
 	register int relocation = 0;
 	register int addr = RELOC_ADDRESS(p);
@@ -1215,8 +1258,8 @@ int reloc_size;
 	register unsigned int mask = 0;
 #else
 	register int length = RELOC_TARGET_SIZE(p);
-#endif	
-	    
+#endif
+
 	if (RELOC_EXTERN_P(p)) {
 	    register int symindex = symbolnum * sizeof (struct nlist);
 
@@ -1231,9 +1274,9 @@ int reloc_size;
 	} else switch (symbolnum) {
 	    case N_TEXT:
 	    case N_TEXT | N_EXT:
-	    	relocation = text_relocation;
+		relocation = text_relocation;
 		break;
-		    
+
 	    case N_DATA:
 	    case N_DATA | N_EXT:
 		/* A word that points to beginning of the the data section
@@ -1242,7 +1285,7 @@ int reloc_size;
 		   the file's text.  */
 		relocation = data_relocation;
 		break;
-		    
+
 	    case N_BSS:
 	    case N_BSS | N_EXT:
 		/* Similarly, an input word pointing to the beginning of
@@ -1250,12 +1293,12 @@ int reloc_size;
 		   of the file.  */
 		relocation = bss_relocation;
 		break;
-		    
+
 	    case N_ABS:
 	    case N_ABS | N_EXT:
 		/* just in case */
 		break;
-		
+
 	    default:
 		break;
 	}
@@ -1276,32 +1319,32 @@ int reloc_size;
 
 	switch (RELOC_TARGET_SIZE(p)) {
 	    case 0:
-	    	*(char *) (data + addr) &= ~mask;
+		*(char *) (data + addr) &= ~mask;
 		*(char *) (data + addr) |= relocation;
 		break;
-		    
+
 	    case 1:
-	    	*(short *) (data + addr) &= ~mask;
+		*(short *) (data + addr) &= ~mask;
 		*(short *) (data + addr) |= relocation;
 		break;
-		    
+
 	    case 2:
-	    	*(long *) (data + addr) &= ~mask;
+		*(long *) (data + addr) &= ~mask;
 		*(long *) (data + addr) |= relocation;
 		break;
 
 	    default: break;
 	}
 
-#else	
-	
+#else
+
 	switch (length) {
 	    case 0: *(char *) (data + addr) += relocation;
 		    break;
-		    
+
 	    case 1: *(short *) (data + addr) += relocation;
 		    break;
-		    
+
 	    case 2: *(int *) (data + addr) += relocation;
 		    break;
 
@@ -1327,8 +1370,8 @@ struct file_entry *entry;
     int text_offset;
 
     /* number of relocation info that describes an external references. */
-    int tr_entry_count = 0, dr_entry_count = 0;	
-				     
+    int tr_entry_count = 0, dr_entry_count = 0;
+
     struct relocation_info *reloc_buf;
 
     if (!entry->header_read_flag)
@@ -1337,7 +1380,7 @@ struct file_entry *entry;
     text_offset = entry->starting_offset + N_TXTOFF (entry->header);
 
     /* For the text segment */
-    
+
     if (entry->header.a_trsize) {
 	reloc_buf = (struct relocation_info *)
 	    _dld_malloc (entry->header.a_trsize);
@@ -1367,7 +1410,7 @@ struct file_entry *entry;
     }
 
     /* For the data segment */
-    
+
     if (entry->header.a_drsize) {
 	reloc_buf = (struct relocation_info *)
 	    _dld_malloc (entry->header.a_drsize);
@@ -1390,7 +1433,7 @@ struct file_entry *entry;
 		_dld_malloc (dr_entry_count * sizeof (struct dld_reloc_info));
 	    do_local_relocation (entry->data_start_address,
 				 entry->data_start_address -
-				 entry->header.a_data, 
+				 entry->header.a_data,
 				 reloc_buf, entry->data_reloc,
 				 entry->header.a_drsize, entry);
 	    free (reloc_buf);
@@ -1405,7 +1448,7 @@ struct file_entry *entry;
 	free (entry->symbols);
 	entry->symbols = 0;
     }
-    
+
 } /* relocate_local_refs */
 
 
@@ -1417,7 +1460,7 @@ struct file_entry *entry;
    RELOC_INFO is the address of the relocation info, in core.
    RELOC_SIZE is its length in bytes.
    REVERSE is true when an 'un-relocation' is to be done.
-*/   
+*/
 
 static void
 perform_relocation (data, pc_relocation, reloc_info, reloc_size, reverse)
@@ -1431,22 +1474,32 @@ int reverse;
 
     struct dld_reloc_info *end
 	= reloc_info + reloc_size / sizeof (struct dld_reloc_info);
-    
+
 #if defined(sun) && defined(sparc)
     if (reverse) return;
 #endif
-    
+
     for (; p < end; p++) {
+#ifdef linux
+	register int relocation;
+	register symbol *sp = p->sp;
+#else
 	register int relocation = p->sp->value;
+#endif
 	register struct relocation_info *r = &(p->reloc_info);
 	register int addr = RELOC_ADDRESS(r);
-	
+
 #if defined(sun) && defined(sparc)
 	register unsigned int mask = 0;
 #else
 	register int length = RELOC_TARGET_SIZE(r);
-#endif	
-	    
+#endif
+
+#ifdef linux
+	while (sp->indirect) sp = sp->indirect;
+	relocation = sp->value;
+#endif
+
 	if (RELOC_PCREL_P(r))
 	    relocation -= pc_relocation;
 
@@ -1463,42 +1516,42 @@ int reverse;
 
 	switch (RELOC_TARGET_SIZE(r)) {
 	    case 0:
-	    	*(char *) (data + addr) &= ~mask;
+		*(char *) (data + addr) &= ~mask;
 		*(char *) (data + addr) |= relocation;
 		break;
-		    
+
 	    case 1:
-	    	*(short *) (data + addr) &= ~mask;
+		*(short *) (data + addr) &= ~mask;
 		*(short *) (data + addr) |= relocation;
 		break;
-		    
+
 	    case 2:
-	    	*(long *) (data + addr) &= ~mask;
+		*(long *) (data + addr) &= ~mask;
 		*(long *) (data + addr) |= relocation;
 		break;
 
 	    default: break;
 	}
 
-#else	
-	
+#else
+
 	if (reverse) relocation = - relocation;
-	    
+
 	switch (length) {
 	  case 0:
 	    if (RELOC_MEMORY_SUB_P(r))
 		*(char *) (data + addr) = relocation - *(char *) (data + addr);
 	    else *(char *) (data + addr) += relocation;
 	    break;
-		    
+
 	  case 1:
 	    if (RELOC_MEMORY_SUB_P(r))
 		*(short *) (data + addr) =
 		    relocation - *(short *) (data + addr);
 	    else *(char *) (data + addr) += relocation;
 	    break;
-		    
-	  case 2: 
+
+	  case 2:
 	    if (RELOC_MEMORY_SUB_P(r))
 		*(int *) (data + addr) = relocation - *(int *) (data + addr);
 	    else *(int *) (data + addr) += relocation;
@@ -1529,7 +1582,7 @@ char *filename;
     entry->filename = entry->local_sym_name =
 	(char *) _dld_malloc (strlen (filename) + 1);
     strcpy (entry->local_sym_name, filename);
-    
+
     if (!ABSOLUTE_FILENAME_P(filename)) {
 	char name[MAXPATHLEN];
 	entry->filename = concat (getwd(name), "/", filename);
@@ -1552,11 +1605,11 @@ int load_text;			    /* used only by dld_init */
 {
     register int len;
     int magicnum;
-    
+
     len = read (desc, &magicnum, sizeof magicnum);
     if (len != sizeof magicnum)
 	fatal (DLD_EBADHEADER);
-    
+
 #ifdef atarist
     if (!load_text || !N_BADMAG (*((struct exec *)&magicnum))) {
 #else
@@ -1570,7 +1623,7 @@ int load_text;			    /* used only by dld_init */
 	if (load_text) read_text_and_data (desc, entry);
     } else {
 	char armag[SARMAG];
-	
+
 	lseek (desc, 0, 0);
 	if (SARMAG != read (desc, armag, SARMAG) || strncmp (armag, ARMAG, SARMAG))
 	    fatal (DLD_EBADOBJECT);
@@ -1597,16 +1650,16 @@ register struct file_entry *entry;
 	entry->bss_start_address = entry->data_start_address +
 	    entry->header.a_data;
     } else entry->data_start_address = entry->bss_start_address = 0;
-    
+
 
     /* Read text and data sections into core.
        Note that the bss segment does not actually take up space in the
        object file, so its size must be subtracted from SIZE */
-    
+
     lseek (desc, entry->starting_offset + N_TXTOFF(entry->header), 0);
     size -= entry->header.a_bss;
-    if (size != read (desc, entry->text_start_address, size)) {
-	free (entry->text_start_address);
+    if (size != read (desc, (char *)entry->text_start_address, size)) {
+	free ((void *)entry->text_start_address);
 	entry->text_start_address = entry->data_start_address =
 	    entry->bss_start_address = 0;
 	fatal (DLD_ENODATA);
@@ -1616,7 +1669,7 @@ register struct file_entry *entry;
     if (entry->header.a_bss)
 	bzero ((void *) entry->bss_start_address, entry->header.a_bss);
 } /* read_text_and_data */
-    
+
 
 /* Allocate memory for the text and data segments and relocate all local
    symbols */
@@ -1626,7 +1679,7 @@ int desc;
 register struct file_entry *entry;
 {
     /* Compute start addresses of each sections and symbols.  */
-    
+
     if (entry->library_flag) {
 	register struct file_entry *subentry = entry->subfiles;
 	for (; subentry; subentry = subentry->chain) {
@@ -1638,8 +1691,8 @@ register struct file_entry *entry;
 	relocate_local_refs (desc, entry);
     }
 } /* relocate_entry_symbols */
-	
- 
+
+
 /*  Find all modules have all external references defined but not resolved. */
 void
 _dld_patch_all_files (entry)
@@ -1686,20 +1739,19 @@ register struct file_entry *entry;
  * reset the executable_flag of the given entry, and then recursively
  * propagate this to all modules that reference symbols in this entry.
  */
-static void
-invalidate (entry)
+static void Invalidate (entry)
 struct file_entry *entry;
 {
     register struct file_chain *p;
-    
+
     if (entry == 0)
 	return;
-    
+
     entry->executable_flag = 0;
 
     for (p = entry->refs_by; p; p = p->next)
 	if (p->entry->executable_flag)
-	    invalidate (p->entry);
+	    Invalidate (p->entry);
 } /* invalidate */
 
 /*
@@ -1722,38 +1774,38 @@ find_all_executable_modules ()
 	}
 	p->executable_flag = p->all_symbols_resolved_flag;
     }
-    
+
     /* invalidate those modules that are not (yet) executable. */
     for (p = _dld_latest_entry; p; p = p->chain) {
 	if (p->library_flag) {
 	    register struct file_entry *q = p->subfiles;
 	    while (q) {
 		if (!q->all_symbols_resolved_flag && q->refs_by)
-		    invalidate (q);
+		    Invalidate (q);
 		q = q->chain;
 	    }
 	} else if (!p->all_symbols_resolved_flag && p->refs_by)
-	    invalidate (p);
+	    Invalidate (p);
     }
     _dld_exec_flags_valid = 1;
 } /* find_all_executable_modules */
 
 
 /* remove all reference pointers *TO* ENTRY */
-static void 
+static void
 remove_cross_references (head_of_chain, entry)
 struct file_entry *head_of_chain, *entry;
 {
     register struct file_entry *ep = head_of_chain;
-    
+
     while (ep) {
 	register struct file_chain *p = ep->refs;
-	
+
 	if (ep->library_flag)
 	    remove_cross_references (ep->subfiles, entry);
 	else {
 	    register struct file_chain *prev = 0;
-	    
+
 	    while (p) {
 		if (p->entry == entry || p->entry->superfile == entry) {
 		    del_link_list_elt (ep->refs, prev, p, next);
@@ -1792,7 +1844,7 @@ struct file_entry *entry;
 
     while (p) {
 	register struct file_chain *next = p->next;
-	
+
 	if (p->entry->ref_count)    /* ref_count may already be set to zero */
 	    (p->entry->ref_count)--;
 	if (p->entry->ref_count == 0)
@@ -1816,7 +1868,7 @@ register struct file_entry *entry;
 	while (subentry)
 	    subentry = kill_entry (subentry);
     }
-	
+
     p = entry->refs;
     while (p) {
 	register struct file_chain *next_chain = p->next;
@@ -1839,12 +1891,12 @@ register struct file_entry *entry;
     if (entry->symbols) free (entry->symbols);
 
     if (entry->strings) free (entry->strings);
-    
+
     if (entry->text_reloc) free (entry->text_reloc);
 
     if (entry->data_reloc) free (entry->data_reloc);
 
-    if (entry->text_start_address) free (entry->text_start_address);
+    if (entry->text_start_address) free ((void *)entry->text_start_address);
 
     {
 	register struct file_entry *next = entry->chain;
@@ -1869,13 +1921,13 @@ clean_up ()
 	for (; subentry; subentry = subentry->chain)
 	    subentry->ref_count = 0;
     }
-	
+
     remove_cross_references (_dld_latest_entry, _dld_latest_entry);
 
     cleanup_obsolete_entries (_dld_latest_entry);
 
     cleanup_symtab ();
-    
+
     _dld_latest_entry = kill_entry (_dld_latest_entry);
 } /* clean_up */
 
@@ -1889,7 +1941,7 @@ struct file_entry *entry;
 {
     register struct file_entry *p = entry;
     register struct file_entry *prev = 0;
-    
+
     if (p == 0) return p;
 
     while (p) {
@@ -1903,9 +1955,9 @@ struct file_entry *entry;
 
 	if (p->ref_count == 0) {
 	    register struct file_entry *next;
-	    
+
 	    next = kill_entry (p);
-	
+
 	    if (prev == 0) {
 		entry = next;
 		p = entry;
@@ -1922,23 +1974,24 @@ struct file_entry *entry;
     return entry;
 } /* do_unlink */
 
+int
 dld_init (myname)
 char *myname;
 {
     int desc;
-    
+
     page_size = getpagesize ();
     bzero (_dld_symtab, TABSIZE * sizeof(symbol *));
     _dld_latest_entry = 0;
     _dld_dummy_entry = 0;
     dld_undefined_sym_count = 0;
     _dld_exec_flags_valid = 0;
-    
+
     if (myname == 0) {
 	dld_errno = DLD_ENOFILE;
 	return dld_errno;
     }
-    
+
     if (setjmp (_dld_env)) {
 	clean_up ();
 	file_close ();
@@ -1974,16 +2027,16 @@ char *myname;
 	free (_dld_latest_entry->symbols);
 	_dld_latest_entry->symbols = 0;
     }
-    
+
     file_close ();
 
     _dld_latest_entry->undefined_symbol_count = 0;
     _dld_latest_entry->all_symbols_resolved_flag = 1;
-    
+
     return 0;
 } /* dld_init */
 
-
+int
 dld_link (object_file)
 char *object_file;
 {
@@ -2000,32 +2053,34 @@ char *object_file;
     dld_errno = 0;
     file_close ();		    /* file might be opened in the last
 				       call */
-    
+
     if (object_file == 0) {
 	dld_errno = DLD_ENOFILE;
 	return dld_errno;
     }
-    
+
     _dld_latest_entry = make_entry (object_file);
 
     desc = file_open (_dld_latest_entry);
-    
+
     read_file_symbols (desc, _dld_latest_entry, 1);
 
     if (_dld_latest_entry->library_flag && _dld_latest_entry->subfiles == 0) {
 	free (_dld_latest_entry->local_sym_name);
 	if (_dld_latest_entry->filename != _dld_latest_entry->local_sym_name)
 	    free (_dld_latest_entry->filename);
-	
+	_dld_latest_entry->filename = NULL;
+	_dld_latest_entry->local_sym_name = NULL;
+
 	_dld_latest_entry = kill_entry (_dld_latest_entry);
 	file_close ();
 	return 0;
     }
-    
+
     relocate_entry_symbols (desc, _dld_latest_entry);
-    
+
     file_close ();
-    
+
     _dld_patch_all_files (_dld_latest_entry);
 
     _dld_exec_flags_valid = 0;
@@ -2039,19 +2094,27 @@ dld_get_bare_symbol (name)
 char *name;
 {
     register symbol *sp;
-    
+
     if (name == 0)
 	return 0;
 
     sp = _dld_getsym_soft (name);
 
+#ifdef linux
+    if (sp) {
+      while (sp->indirect) sp = sp->indirect;
+      if (sp->defined)
+	  return sp->value;
+    }
+#else
     if (sp)
 	if (sp->defined)
 	    return sp->value;
+#endif
     return 0;
 } /* dld_get_bare_symbol */
 
-	
+
 /* given a file_entry, unlink that (and all its decendents).  Modules still
    referenced by the remainings will not be unlinked.
    ENTRY is assumed to be a valid pointer to a file_entry structure.
@@ -2094,6 +2157,7 @@ int force;
 /*
  * return true if the named function can be safely exeucted.
  */
+int
 dld_function_executable_p (name)
 char name[];
 {
@@ -2113,16 +2177,31 @@ char name[];
     sp = _dld_getsym_soft (p);
     free (p);
 
+#ifdef linux
+    if (sp && (sp->defined == (N_EXT | N_TEXT) ||
+	       sp->defined == (N_EXT | N_INDR))) {
+#else
     if (sp && sp->defined == (N_EXT | N_TEXT)) {
+#endif
 	register struct file_entry *fe = sp->defined_by;
 
 	if (fe == 0) return 0;
-	
+
+#ifdef linux
+	while (sp->indirect) sp = sp->indirect;
+	if (sp->defined != (N_EXT | N_TEXT)) return 0;
+#endif
+
 	if (!_dld_exec_flags_valid)
 	    find_all_executable_modules ();
 
 	return fe->executable_flag;
     }
 
+#ifdef linux
+    if (sp && sp->defined == (N_EXT | N_ABS))
+	return 1;
+#endif
+
     return 0;
 } /* dld_function_executable_p */
