@@ -1,7 +1,7 @@
 /* get_func.c -- return the entry point of the given function. */
 
 /* This file is part of DLD, a dynamic link/unlink editor for C.
-   
+
    Copyright (C) 1990 by W. Wilson Ho.
 
    The author can be reached electronically by how@cs.ucdavis.edu or
@@ -30,7 +30,7 @@ char name[];
 
     if (name == 0)
 	return 0;
-    
+
     if (setjmp (_dld_env))
 	return 0;
 
@@ -42,9 +42,17 @@ char name[];
     sp = _dld_getsym_soft (p);
     free (p);
 
+#ifdef linux
+    if (sp) {
+	while (sp->indirect) sp = sp->indirect;
+	if ((sp->defined == (N_EXT | N_TEXT)) || (sp->defined == (N_EXT | N_ABS)))
+	    return sp->value;
+    }
+#else
     if (sp)
 	if (sp->defined == (N_EXT | N_TEXT))
 	    return sp->value;
+#endif
     return 0;
 } /* dld_get_func */
 
